/*  compensatory_synapse.h
 *  This file is part of NEST.
    Chris built it.  It changes the weight based on the firing behaviour of the
    pre and post-synaptic neurons, and the total outgoing synaptic weight
    of the pre-synaptic neuron.

    The constants can't be set from PyNN.  It seems to work pretty well.  
 *
 */
 
#ifndef COMPENSATORY_SYNAPSE_H
#define COMPENSATORY_SYNAPSE_H

// Includes from nestkernel:
#include "connection.h"
#include "compense_iaf_cond_exp.h"
#include <math.h>  //undone is there a nest version of this?

/* BeginDocumentation
  Name: compensatory_synapse - undone

  Description:
  undone

  Transmits: SpikeEvent

  Remarks: undone

  SeeAlso: synapsedict
*/

namespace mynest
{

/**
 * undone 
 *
 * For a discussion of how synapses are created and represented in NEST 2.6,
 * please see Kunkel et al, Front Neuroinform 8:78 (2014), Sec 3.3.
 */
template < typename targetidentifierT >
class CompensatorySynapse : public nest::Connection< targetidentifierT >
{
private:
  double weight_; //!< Synaptic weight
  double Wmax_; //target weight
  double min_weight_;
  double time_window_;
  double learning_rate_;
  size_t  preSynapticCellID;  
  nest::Node* preSynapticCell;  //This gets set in check_connection.
                                //So, the constructor fails if it
                                //references it, because it check_connection
                                //has not been called.


public:
  //! Type to use for representing common synapse properties
  typedef nest::CommonSynapseProperties CommonPropertiesType;

  //! Shortcut for base class
  typedef nest::Connection< targetidentifierT > ConnectionBase;

  // Explicitly declare all methods inherited from the dependent base
  // ConnectionBase. This avoids explicit name prefixes in all places these
  // functions are used. Since ConnectionBase depends on the template parameter,
  // they are not automatically found in the base class.
  using ConnectionBase::get_target;

  /**
   * Default Constructor.
   * Sets default values for all parameters. Needed by GenericConnectorModel.
   */
  CompensatorySynapse()
    : ConnectionBase()
    , weight_( 1.0 )
    , Wmax_ ( 2.0 ) //target weight for 100% 
    , min_weight_ ( 0.001 )
    , time_window_( 10.0 )
    , learning_rate_( 0.01 )
  {
    //You'd like to set increase the pif currentSynapticWeight here, but can't
    //because the check_connection has not yet been called, so this synapse
    //currently has no presynaptic neuron.  So, save the initial weight 
    //here and add it to the presynaptic weight in check_connection
    //((compense_iaf_cond_exp*) preSynapticCell)->setCurrentSynapticWeight(weight_);
  }

  //! Default Destructor.
  ~CompensatorySynapse()
  {
  }

  /**
   * Helper class defining which types of events can be transmitted.
   *
   * These methods are only used to test whether a certain type of connection
   * can be created.
   *
   * `handles_test_event()` should be added for all event types that the
   * synapse can transmit. The methods shall return `invalid_port_`; the
   * return value will be ignored.
   *
   * Since this is a synapse model dropping spikes, it is only for spikes,
   * therefore we only implement `handles_test_event()` only for spike
   * events.
   *
   * See Kunkel et al (2014), Sec 3.3.1, for background information.
   */
  class ConnTestDummyNode : public nest::ConnTestDummyNodeBase
  {
  public:
    using nest::ConnTestDummyNodeBase::handles_test_event;
    nest::port
    handles_test_event( nest::SpikeEvent&, nest::rport )
    {
      return nest::invalid_port_;
    }

    nest::port
    handles_test_event( nest::DSSpikeEvent&, nest::rport )
    {
      return nest::invalid_port_;
    }
  };

  /**
   * Check that requested connection can be created.
   *
   * This function is a boilerplate function that should be included unchanged
   * in all synapse models. It is called before a connection is added to check
   * that the connection is legal. It is a wrapper that allows us to call
   * the "real" `check_connection_()` method with the `ConnTestDummyNode
   * dummy_target;` class for this connection type. This avoids a virtual
   * function call for better performance.
   *
   * @param s  Source node for connection
   * @param t  Target node for connection
   * @param receptor_type  Receptor type for connection
   * @param lastspike Time of most recent spike of presynaptic (sender) neuron,
   *                  not used here
   */
  void
  check_connection( nest::Node& s,
    nest::Node& t,
    nest::rport receptor_type,
    double t_lastspike,
    const CommonPropertiesType& )
  { 
    //set the presynaptic cell when this is called.
    preSynapticCellID = s.get_gid();
    preSynapticCell = &s;
    //undone assumes check_connection is only called once per synapse.
    //Add the weight of this synapse to the presynaptic weight
    double preSynapticNeuronCurrentWeight = ((nest::compense_iaf_cond_exp*) preSynapticCell)->getCurrentSynapticWeight();
    preSynapticNeuronCurrentWeight =  preSynapticNeuronCurrentWeight + weight_;
    ((nest::compense_iaf_cond_exp*) preSynapticCell)->setCurrentSynapticWeight(preSynapticNeuronCurrentWeight);

    ConnTestDummyNode dummy_target;
    ConnectionBase::check_connection_( dummy_target, s, t, receptor_type );
    
  }

  /**
   * Send an event to the receiver of this connection.
   * @param e The event to send
   * @param t Thread
   * @param t_lastspike Point in time of last spike sent.
   * @param cp Common properties to all synapses.
   */
  void send( nest::Event& e,
    nest::thread t,
    double t_lastspike,
    const CommonPropertiesType& cp );

  // The following methods contain mostly fixed code to forward the
  // corresponding tasks to corresponding methods in the base class and the w_
  // data member holding the weight.

  //! Store connection status information in dictionary
  void get_status( DictionaryDatum& d ) const;

  /**
   * Set connection status.
   *
   * @param d Dictionary with new parameter values
   * @param cm ConnectorModel is passed along to validate new delay values
   */
  void set_status( const DictionaryDatum& d, nest::ConnectorModel& cm );

  //! Allows efficient initialization on contstruction
  void
  set_weight( double w )
  {
    weight_ = w;
  }

//correlatory rule from lots of places including 
//huyck 2007 creating hierarchical categories using CAs
//D+ = 1-wt
//D- = wt*-1
 double getCorrelatoryChange(double weight, bool coFire) {
   //currentCorrelation = (target_weight_-weight) / target_weight
   if (coFire)  {
     //return  1.0 - weight; //if target_weight_ = 1.0
     return (this->Wmax_ - weight);
   }
   else {
     return weight*-1.0;
   }
 }

//compensatory modifier from huyck 2007 creating hierarchical categories 
//using CAs
//M+ = 10 ^ WB-Wj = 10^(targetWeight-currentWeight) or 2^
//M- = 2^(currentWeight - targetWeight)
double getCompensatoryModifier(double currentWeight, double targetWeight, 
				bool coFire) {
  double weightDiff;
  if (coFire) {
    weightDiff = targetWeight - currentWeight;
  }
  else {
    weightDiff = currentWeight-targetWeight;
  }
  double modifier = pow(2,weightDiff);
  return modifier;
  }
};

template < typename targetidentifierT >
inline void
CompensatorySynapse< targetidentifierT >::send( nest::Event& e,
  nest::thread t,
  double t_lastSpike,
  const CommonPropertiesType& props )
{
  double t_spike = e.get_stamp().get_ms();
  nest::Node* postSynapticCell = get_target( t );

  //Leave the weight unchanged if this is the first time this synapse used.
  if(t_lastSpike >0)  {
    //shamelessly trying to take code from stdp_connection.h
    //Get the last time the pre-synaptic neuron fired
    //See if the post-synaptic neuron fired just after it.
    //Get spike history in relevant range (the last time we registered (t1)
    //to t1+time_window_) from post-synaptic neuron
    std::deque< nest::histentry >::iterator startPostSpikes;
    std::deque< nest::histentry >::iterator finishPostSpikes;
    //this won't work if they fire in the same cycle
    //postSynapticCell->get_history(t_lastSpike,t_lastSpike + time_window_, 
    //			  &startPostSpikes, &finishPostSpikes);
    postSynapticCell->get_history(t_lastSpike-1,t_lastSpike + time_window_, 
    			  &startPostSpikes, &finishPostSpikes);

    //if there was a spike between last and last + time_window_ they've co-fired
    bool fired = (startPostSpikes != finishPostSpikes);
    bool coFire = fired;
    if (coFire) {
      coFire = (startPostSpikes->t_ < (t_lastSpike + time_window_));
    }
    double baseWeightChange = getCorrelatoryChange(weight_,coFire);

    //Get the values from the preSynapticNeuron
    double preSynapticNeuronTargetWeight = ((nest::compense_iaf_cond_exp*) preSynapticCell)->getTargetSynapticWeight();
    double preSynapticNeuronCurrentWeight = ((nest::compense_iaf_cond_exp*) preSynapticCell)->getCurrentSynapticWeight();

    double compensatoryModifier = getCompensatoryModifier(preSynapticNeuronCurrentWeight,preSynapticNeuronTargetWeight,coFire);

    //Cap the change to the max change, then use learning_rate.
    double baseChange = (baseWeightChange*compensatoryModifier);
    if (baseChange > 0) {
      if ((baseChange + weight_) > Wmax_) {
	baseChange = Wmax_ - weight_;
      }
    }
    else
      if ((baseChange + weight_) < min_weight_) {
	baseChange = min_weight_ - weight_;
      }

    //This doesn't cap before learning_rate
    //double newWeight = weight_ + (baseWeightChange*learning_rate_*compensatoryModifier);
    double newWeight = weight_ + (baseChange*learning_rate_);
    if (newWeight > Wmax_) newWeight = Wmax_;
    if (newWeight < min_weight_) newWeight = min_weight_;
    double weightChange = newWeight-weight_;
    set_weight (newWeight);

    ((nest::compense_iaf_cond_exp*) preSynapticCell)->setCurrentSynapticWeight(preSynapticNeuronCurrentWeight+weightChange);
  }

  // Even time stamp, we send the spike using the normal sending mechanism
  // send the spike to the target
  e.set_weight( weight_ );
  e.set_delay( ConnectionBase::get_delay_steps() );
  e.set_receiver( *ConnectionBase::get_target( t ) );
  e.set_rport( ConnectionBase::get_rport() );
  e(); // this sends the event

  //This tells the post-synaptic neuron to record firings from now (t_spike)
  postSynapticCell->register_stdp_connection( t_spike );
}

template < typename targetidentifierT >
void
CompensatorySynapse< targetidentifierT >::get_status(
  DictionaryDatum& d ) const
{
  ConnectionBase::get_status( d );
  def< double >( d, nest::names::weight, weight_ );
  def< long >( d, nest::names::size_of, sizeof( *this ) );
}

template < typename targetidentifierT >
void
CompensatorySynapse< targetidentifierT >::set_status(
  const DictionaryDatum& d,
  nest::ConnectorModel& cm )
{
  ConnectionBase::set_status( d, cm );
  updateValue< double >( d, nest::names::weight, weight_ );
}

} // namespace

#endif // compensatory_synapse.h
